/*
Gamespy NAT negotiation plugin for Proxocket 0.1
by Luigi Auriemma
e-mail: aluigi@autistici.org
web:    aluigi.org

You need Proxocket:
http://aluigi.org/mytoolz.htm#proxocket

How to use:
- put Proxocket's ws2_32.dll in the folder of the main
  program/tool to which you want to apply gsnatneg (for
  example where is located one of my fake players tool just
  to test how it works)
- put this myproxocket.dll in the same folder
- use the program/tool normally
- in case of problems or doubts read the Proxocket's manual

Note that this solution will apply gsnatneg to all the UDP
streams of the main program!
So use it only if you are sure that the host you want to
reach is behind NAT and uses the Gamespy NAT negotiation.

How to compile:
  gcc -o myproxocket.dll myproxocket.c -shared 
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <winsock.h>
#include <windows.h>



static HMODULE wsock = NULL;
static WINAPI SOCKET (*real_socket)(int af, int type, int protocol) = NULL;
static WINAPI int (*real_connect)(SOCKET s, const struct sockaddr *name, int namelen) = NULL;
static WINAPI int (*real_closesocket)(SOCKET s) = NULL;
static WINAPI int (*real_recv)(SOCKET s, char *buf, int len, int flags) = NULL;
static WINAPI int (*real_recvfrom)(SOCKET s, char *buf, int len, int flags, struct sockaddr *from, int *fromlen) = NULL;
static WINAPI int (*real_send)(SOCKET s, char *buf, int len, int flags) = NULL;
static WINAPI int (*real_sendto)(SOCKET s, char *tbuf, int len, int flags, const struct sockaddr *to, int tolen) = NULL;
static WINAPI struct hostent* (*real_gethostbyname)(const char *name) = NULL;



void init_myproxocket(void) {
    char    winpath[MAX_PATH];

    if(wsock) return;   // already set

    GetSystemDirectory(winpath, sizeof(winpath));
    strcat(winpath, "\\ws2_32.dll");

    wsock = LoadLibrary(winpath);
    if(!wsock) return;

    real_socket         = (void *)GetProcAddress(wsock, "socket");
    real_connect        = (void *)GetProcAddress(wsock, "connect");
    real_closesocket    = (void *)GetProcAddress(wsock, "closesocket");
    real_recv           = (void *)GetProcAddress(wsock, "recv");
    real_recvfrom       = (void *)GetProcAddress(wsock, "recvfrom");
    real_send           = (void *)GetProcAddress(wsock, "send");
    real_sendto         = (void *)GetProcAddress(wsock, "sendto");
    real_gethostbyname  = (void *)GetProcAddress(wsock, "gethostbyname");
}



void free_myproxocket(void) {
    if(wsock) {
        FreeLibrary(wsock);
        wsock = NULL;
    }
}



#define socket          real_socket
#define connect         real_connect
#define closesocket     real_closesocket
#define recv            real_recv
#define recvfrom        real_recvfrom
#define send            real_send
#define sendto          real_sendto
#define gethostbyname   real_gethostbyname

#include "gsnatneg.c"



int __cdecl mysendto(SOCKET s, u_char **retbuf, int len, int flags, const struct sockaddr *to, int tolen) {
    static SOCKET   old_sd      = -1;
    static uint16_t old_port    = 0;

    static int  skip    = 0;    // useful if we don't want to use the real_* solution
    int         ret;

    if(skip) return(len);
    if(!to) return(len);

    // sorry but at the moment I don't know how to avoid
    // gethostbyname using this sendto so I simply skip all
    // the packets versus port 53... yeah lame
    if(ntohs(((struct sockaddr_in *)to)->sin_port) == 53) return(len);

    // if socket is different than the old one (doesn't happen on Windows
    // or the bound port is different then call gsnatneg
    uint16_t    port = 0;
    gsnatneg_get_sock_ip_port(s, &port, NULL, 0);
    if((s != old_sd) || (port != old_port)) {
        //gsnatneg_verbose = 1;

        skip = 1;
        ret = gsnatneg(
            s,              // socket
            "battlefield2", // gamename (should work with almost all the games)
            NULL,           // gamekey
            NULL,           // string IP
            INADDR_NONE,    // IP
            0,              // port
            (void *)to,     // IP:port
            NULL);          // callback
        skip = 0;

        old_sd = s;
        gsnatneg_get_sock_ip_port(s, &port, NULL, 0);
        if(port) old_port = port;

        if(ret < 0) {
            return(len);    // ignore the errors, maybe the host doesn't use gsnatneg

            //printf("\nError: gsnatneg failed at line %d\n", ret);
            //exit(1);
            //return(SOCKET_ERROR);
        }
    }
    return(len);
}



BOOL WINAPI DllMain(HINSTANCE hinstDLL, DWORD fdwReason, LPVOID lpReserved) {
    switch(fdwReason) {
        case DLL_PROCESS_ATTACH: {
            DisableThreadLibraryCalls(hinstDLL);
            init_myproxocket();
            break;
        }
        case DLL_PROCESS_DETACH: {
            free_myproxocket();
            break;
        }
        default: break;
    }
    return(TRUE);
}


